==> MeatAxe-2.3.0.tar.gz: src/precond.c <==
/* ============================= C MeatAxe ==================================
   File:        $Id: precond.c,v 1.13 1998/06/25 19:02:34 mringe Exp $
   Comment:     This program creates various files needed by the tensor
		condense programs.
   --------------------------------------------------------------------------
   Written by Michael Ringe.
   (C) Copyright 1998  Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"

#include <ctype.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>


/* --------------------------------------------------------------------------
   Global data
   -------------------------------------------------------------------------- */

char TkiName[LAT_MAXBASENAME];	/* Base name for .tki file */
TK_Info TKInfo;			/* Data from .tki file */
Lat_Info InfoM, InfoN;		/* Data from .cfinfo files */
int CfIsLinked[LAT_MAXCF];	/* Constituent (of N) is already linked */
matrix_t *Trans[LAT_MAXCF];	/* */
int opt_s = 0;			/* Do NOT use standard generators */

static char *helptext[] = {
"SYNTAX",
"    precond [-s] <info> <M> <N>",
"",
"OPTIONS",
/*
"    -s    Do not read generators in standard basis. Use this option if you",
"          have used symbas instead of pwkond",
*/
"",
"FILES",
"    <M>.cfinfo           i  Info file for M (produced by chop/pwkond)",
"    <N>.cfinfo           i  Info file for N (produced by chop/pwkond)",
"    <M/N>1a.std.1 ...    i  Standard generators for each constituent",
"    <info>.tki           o  Tensor condense info file",
"    <info>.q.1 ...       o  Basis matrices for 'tensor condense'",
"    <info>.p.1 ...       o  Projection matrices for 'tensor condense'",
"",
"DESCRIPTION",
"    This program produces various files and information needed for the",
"    tensor condensation (tkond):",
"    1) The compositionfactors in <m>.cfinfo and <n>.cfinfo are compared and",
"       sorted if necessary and some information is written to <condinfo>.",
"    2) The idempotent e belonging to H is calculated in the tensorproduct t.j",
"       of each irreducible with its contragredient representation.",
"    3) t.j(e) is decomposed in basismatrix <q>.j and projectionmatrix <p>.j.",
"",
NULL};

static proginfo_t pinfo =
   { "precond", "Precondensation", "$Revision: 1.13 $", helptext };




/* --------------------------------------------------------------------------
   Init() - Initialize everything

   Description:
     This function initializes global variables, processes command line
     options and arguments, and reads the .cfinfo and .tki files.

   Arguments:
     <argc>: Number of arguments
     <argv>: Command line arguments
   -------------------------------------------------------------------------- */

static void Init(int argc, char *argv[])

{
    int i;
    long fl1, nor1, noc1, fl2, nor2, noc2;
    char fn[LAT_MAXBASENAME+20];

    /* Initialize global data
       ---------------------- */
    memset(&TKInfo,0,sizeof(TKInfo));
    memset(TkiName,0,sizeof(TkiName));
    memset(CfIsLinked,0,sizeof(CfIsLinked));

    /* Initialize the MeatAxe library
       ------------------------------ */
    MtxInit();

    /* Process command line options
       ---------------------------- */
    MtxSetCommandLine(argc,argv,&pinfo);
    while ((i = MtxGetOpt("")) != OPT_END)
    {
	switch (i)
	{
	    case 's': opt_s = 1; break;
	    default:
		MtxErrorExit("Unknown option '%c'",i);
		break;
	}
    }

    /* Process remaining command line arguments
       ---------------------------------------- */
    if (opt_ind != argc - 3)
	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
    strcpy(TkiName,argv[opt_ind]);

    /* Read info files
       --------------- */
    if (Lat_ReadInfo(&InfoM,argv[opt_ind + 1]) != 0)
	MtxErrorExit("Error reading %s.cfinfo",argv[opt_ind + 1]);
    if (Lat_ReadInfo(&InfoN,argv[opt_ind + 2]) != 0)
	MtxErrorExit("Error reading %s.cfinfo",argv[opt_ind + 2]);



    /* Initialize the TKInfo structure
       ------------------------------- */
    TKInfo.NGen = InfoM.NGen;
    TKInfo.NCf = 0;
    TKInfo.Order = 1;
    for (i = 0; i < LAT_MAXCF; ++i)
	TKInfo.CfIndex[0][i] = TKInfo.CfIndex[1][i] = -1;

    /* Some additional checks on input files
       ------------------------------------- */
    if (InfoM.Field != InfoN.Field)
    {
	MtxErrorExit("Incompatible representations: %s is over GF(%d), "
	   "%s is over GF(%d)",InfoM.BaseName,InfoM.Field,InfoN.BaseName,
	   InfoN.Field);
    }
    if (InfoM.NGen != InfoN.NGen)
    {
	MtxErrorExit("Incompatible representations: %s has %d generators, "
	   "%s has %d generators",InfoM.BaseName,InfoM.NGen,InfoN.BaseName,
	   InfoN.NGen);
    }

    /* Print start message
       ------------------- */
    sprintf(fn,"%s.1",InfoM.BaseName);
    fclose(zreadhdr(fn,&fl1,&nor1,&noc1));
    sprintf(fn,"%s.1",InfoN.BaseName);
    fclose(zreadhdr(fn,&fl2,&nor2,&noc2));
    MESSAGE(0,("Beginning pre-condensation of dimension %ld x %ld = %ld\n",
	nor1,nor2,nor1*nor2));
}



/* --------------------------------------------------------------------------
   IsDual() - Find matching constituent

   Description:
     This function decides is a given contituent of M is dual to a given 
     contituent of N.

   Arguments:
     <minfo>: Pointer to <CfInfo> structure for the constituent of M.
     <mgen>: Generators of the constituent of M.
     <nj>: Index of constituent of N.

   Return:
     1 = the constituents are dual, 0 = they are not dual.
   -------------------------------------------------------------------------- */

static int IsDual(CfInfo *minfo, matrix_t *mgen[], int nj)

{
    matrix_t *ngen[MAXGEN];	/* Generators of constituent */
    int j;
    int result;

    /* First check: Dimensions and splitting field must match
       ------------------------------------------------------ */
    if (InfoN.Cf[nj].dim != minfo->dim || InfoN.Cf[nj].spl != minfo->spl)
	return 0;

    /* Read the (contragrediate) generators and compare
       ------------------------------------------------ */
    MESSAGE(2,(" (%s%s)",InfoN.BaseName,Lat_CfName(&InfoN,nj)));
    Lat_ReadCfGens(&InfoN,nj,ngen,LAT_RG_INVERT|LAT_RG_TRANSPOSE);
    result = IsIsomorphic(InfoM.NGen,(const matrix_t **)mgen,minfo,
	(const matrix_t **)ngen,Trans + TKInfo.NCf);
    for (j = 0; j < InfoN.NGen; j++)
	MatFree(ngen[j]);

    return result;
}




/* --------------------------------------------------------------------------
   InitFindConstituentInN() - Find matching constituent

   Description:
     This function finds, for a given constituent of M, the corresponding
     constituent in N (if any exists). The generators in <mgen> are supposed
     to be in standard basis with respect to the constituent's idword.

   Arguments:
     <mj>: Index of the constituent of M.
     <mgen>: Generators of the constituent of M.

   Return:
     Index of corresponding constituent of N, or -1 if not found.
   -------------------------------------------------------------------------- */

static int FindConstituentInN(int mj, matrix_t *mgen[])

{
    int nj;

    /* Main loop: for each constituent of N
       ------------------------------------ */
    for (nj = 0; nj < InfoN.NCf; ++nj)
    {
	/* Discard constituents which are already linked
	   --------------------------------------------- */
	if (CfIsLinked[nj])
	   continue;

	/* Compare with <nj>-th constituent of N
	   ------------------------------------- */
	if (IsDual(InfoM.Cf + mj,mgen,nj))
	{
	    CfIsLinked[nj] = 1;	    /* Mark as linked */
	    return nj;
	}
    }

    /* Corresponding constituent was not found
       --------------------------------------- */
    return -1;
}



/* --------------------------------------------------------------------------
   MkEndo() - Calculate endomorphism ring

   Description:
     This function calculates, for a given irreducible constituent, a
     basis for the endomoprhism ring.

   Arguments:
     <ngen>: Number of generators
     <gen>: Generators.
     <m>: Constituent index
     <cf>: Pointer to constituent information.
     <endo>: Pointer to an array where the basis of End(V) is stored.
     <maxendo>: Maximum number of matrices that can be stored in <endo>.
     <basis>: The 'E-basis' is stored here
   -------------------------------------------------------------------------- */

static void MkEndo(int ngen, const matrix_t **gen, int m, const CfInfo *cf,
    matrix_t **endo, int maxendo)
{
    matrix_t *pw, *nsp;
    wgdata_t *wg;
    int i;
    MTXCONTEXT(MkEndo);

    MTX_ASSERT(maxendo >= cf->spl);

    /* Make the peak word kernel
       ------------------------- */
    wg = WgInit(ngen,gen);
    pw = WgMakeWord(wg,cf->peakword);
    WgFree(wg);
    nsp = MatNullSpace__(MatInsert(pw,cf->peakpol));
    MTX_ASSERT(nsp->nor == cf->spl);
    MatFree(pw);

    /* Calculate a basis of the the endomorphism ring
       ---------------------------------------------- */
    i = MakeEndomorphisms(ngen,gen,nsp,endo);
    MTX_ASSERT(i == 0);

    MatFree(nsp);
}




/* --------------------------------------------------------------------------
   MakeQ() - Calculate the Q matrix

   Description:
     This function calculates the embedding of the condensed tensor product
     (VxW)e into the tensor product VxW for one irreducible constituent.

   Arguments:
     <n>: Constituent index.
     <spl>: Splitting field degree.
     <endo>: Basis of endomorphism ring.
   -------------------------------------------------------------------------- */

static void MakeQ(int n, int spl, const matrix_t **endo)
{		  
    int i;
    int dim = endo[0]->nor;
    matrix_t *q = MatAlloc(endo[0]->fl,spl,dim*dim);
    char fn[200];
    for (i = 0; i < spl; ++i)
    {
	int j;
	matrix_t *y = MatInv(Trans[n]), *x;
	MatMul(y,endo[i]);
	x = MatTr(y);
	MatFree(y);
	for (j = 1; j <= dim; ++j)
	    MatCopyRegion(q,i+1,(j-1) * dim+1,x,j,1,1,-1);
	MatFree(x);
    }
    sprintf(fn,"%s.q.%d",TkiName,n+1);
    MESSAGE(2,("Writing %s\n",fn));
    MatSave(q,fn);
    MatFree(q);
}



/* --------------------------------------------------------------------------
   MakePQ() -

   Description:

   Arguments:
     <n>: Constituent index
     <mj>: Index of corresponding constituent of M.
     <nj>: Index of corresponding constituent of N.
   -------------------------------------------------------------------------- */

#define MAXENDO 20

static void MakePQ(int n, int mj, int nj)

{
    matrix_t *mgen[MAXGEN], *estar[MAXENDO], *endo[MAXENDO], *e, *ei;
    char fn[200];
    int dim = InfoM.Cf[mj].dim;
    int spl = InfoM.Cf[mj].spl;
    int i;
    matrix_t *p;
    PTR pptr;

    MESSAGE(1,("Condensing %s%s x ",InfoM.BaseName,Lat_CfName(&InfoM,mj)));
    MESSAGE(1,("%s%s, [E:k]=%d\n",InfoN.BaseName,Lat_CfName(&InfoN,nj),spl));

    /* Read the generators for the constituent of M
       -------------------------------------------- */
    Lat_ReadCfGens(&InfoM,mj,mgen,LAT_RG_STD);

    /* Make the endomorphism ring
       -------------------------- */
    MESSAGE(2,("Calculating endomorphism ring\n"));
    MkEndo(InfoM.NGen,(const matrix_t **)mgen,mj,InfoM.Cf + mj,endo,MAXENDO);

    /* Calculate the Q matrix
       ---------------------- */
    MESSAGE(2,("Calculating embedding of E\n"));
    MakeQ(n,spl,(const matrix_t **)endo);
    
    /* Calculate the E* matrices
       Note: We should use the symmetry under i<-->k here!
       --------------------------------------------------- */
    MESSAGE(2,("Calculating projection on E\n"));
    e = MatAlloc(zfl,spl,spl);
    pptr = e->d;
    for (i = 0; i < spl; ++i)
    {
	int k;
	for (k = 0; k < spl; ++k)
	{
	    FEL f;
	    matrix_t *x = MatDup(endo[i]);  
	    MatMul(x,endo[k]);
	    f = MatTrace(x);
	    zinsert(pptr,k+1,f);
	    MatFree(x);
	}
	zsetlen(spl);
	zadvance(&pptr,1);
    }
    ei = MatInv(e);
    MatFree(e);

    for (i = 0; i < spl; ++i)
    {
	int k;
	PTR p;
	estar[i] = MatAlloc(zfl,dim,dim);
	p = ei->d;
	zsetlen(spl);
	zadvance(&p,i);
	for (k = 0; k < spl; ++k)
	    MatAddMul(estar[i],endo[k],zextract(p,k+1));
    }
    MatFree(ei);

    /* Calculate the P matrix
       ---------------------- */
    p = MatAlloc(zfl,dim*dim,spl);
    pptr = p->d;
    for (i = 1; i <= dim; ++i)
    {
	int j;
	for (j = 1; j <= dim; ++j)
	{
	    int r;
	    matrix_t *x = MatAlloc(zfl,dim,dim);
	    MatCopyRegion(x,1,i,Trans[n],1,j,dim,1);
	    for (r = 0; r < spl; ++r)
	    {
		FEL f;
		matrix_t *z = MatDup(x);
		MatMul(z,estar[r]);
		f = MatTrace(z);
		zinsert(pptr,r+1,f);
		MatFree(z);
	    }
	    MatFree(x);
	    zsetlen(spl);
	    zadvance(&pptr,1);
	}
    }

    sprintf(fn,"%s.p.%d",TkiName,n+1);
    MESSAGE(2,("Writing %s\n",fn));
    MatSave(p,fn);

    /* Clean up
       -------- */
    MatFree(p);
    for (i = 0; i < spl; ++i)
        MatFree(endo[i]);
}




/* --------------------------------------------------------------------------
   CalcDim() - Calculate dimension of condensed module

   Description:
     This function calculates the dimension of the condensed tensor product.

   Remarks:
     The dimension is stored into <TKInfo.Dim>.
   -------------------------------------------------------------------------- */

static void CalcDim()

{
    int i;

    TKInfo.Dim = 0;
    for (i = 0; i < TKInfo.NCf; ++i)
    {
	int m = TKInfo.CfIndex[0][i];
	int n = TKInfo.CfIndex[1][i];
	TKInfo.Dim += InfoM.Cf[m].mult * InfoN.Cf[n].mult * InfoM.Cf[m].spl;
    }
    MESSAGE(0,("Dimension of condensed module = %d\n",TKInfo.Dim));
}


/* --------------------------------------------------------------------------
   main() - Program entry point
   -------------------------------------------------------------------------- */

int main(int argc, char *argv[])

{
    int mj;

    Init(argc,argv);			/* Initialize everything */

    /* Main loop: for all constituents of M
       ------------------------------------ */
    for (mj = 0; mj < InfoM.NCf; mj++)
    {
	matrix_t *mgen[MAXGEN];	    /* Generators for const. of M */
	int k;
	int nj;

	MESSAGE(0,("Examining %s%s ...",InfoM.BaseName,Lat_CfName(&InfoM,mj)));

	/* Read the generators for the <mj>-th contituent of M, and find
	   the corresponting (=contragredient) constituent in N
	   ------------------------------------------------------------- */
	Lat_ReadCfGens(&InfoM,mj,mgen,0);
	nj = FindConstituentInN(mj,mgen);

	/* Calculate the P and Q matrix for this constituent
	   ------------------------------------------------- */
	if (nj >= 0)
	{
	    MESSAGE(0,(" ... corresponds to %s%s\n",InfoN.BaseName,
		Lat_CfName(&InfoN,nj)));
	    TKInfo.CfIndex[0][TKInfo.NCf] = mj;
	    TKInfo.CfIndex[1][TKInfo.NCf] = nj;
	    MakePQ(TKInfo.NCf,mj,nj);
	    TKInfo.NCf++;
	}
	else
	    MESSAGE(0,(" not found in %s\n",InfoN.BaseName));

	/* Clean up
	   -------- */
	for (k = 0; k < InfoM.NGen; ++k)
	    MatFree(mgen[k]);
    }

    CalcDim();				/* Calculate dimension */
    TK_WriteInfo(&TKInfo,TkiName);	/* Write .tki file */
    MtxCleanup();
    return 0;
}

==> MeatAxe-2.3.1.tar.gz: src/precond.c <==
/* ============================= C MeatAxe ==================================
   File:        $Id: precond.c,v 1.16 1998/09/17 17:03:55 mringe Exp $
   Comment:     This program creates various files needed by the tensor
		condense programs.
   --------------------------------------------------------------------------
   Written by Michael Ringe.
   (C) Copyright 1998  Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"

#include <ctype.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>


/* --------------------------------------------------------------------------
   Global data
   -------------------------------------------------------------------------- */

char TkiName[LAT_MAXBASENAME];	/* Base name for .tki file */
TK_Info TKInfo;			/* Data from .tki file */
Lat_Info InfoM, InfoN;		/* Data from .cfinfo files */
int CfIsLinked[LAT_MAXCF];	/* Constituent (of N) is already linked */
matrix_t *Trans[LAT_MAXCF];	/* */
int opt_s = 0;			/* Do NOT use standard generators */

static char *helptext[] = {
"SYNTAX",
"    precond [-s] <info> <M> <N>",
"",
"OPTIONS",
/*
"    -s    Do not read generators in standard basis. Use this option if you",
"          have used symbas instead of pwkond",
*/
"",
"FILES",
"    <M>.cfinfo           i  Info file for M (produced by chop/pwkond)",
"    <N>.cfinfo           i  Info file for N (produced by chop/pwkond)",
"    <M/N>1a.std.1 ...    i  Standard generators for each constituent",
"    <info>.tki           o  Tensor condense info file",
"    <info>.q.1 ...       o  Basis matrices for 'tensor condense'",
"    <info>.p.1 ...       o  Projection matrices for 'tensor condense'",
"",
"DESCRIPTION",
"    This program produces various files and information needed for the",
"    tensor condensation (tkond):",
"    1) The compositionfactors in <m>.cfinfo and <n>.cfinfo are compared and",
"       sorted if necessary and some information is written to <condinfo>.",
"    2) The idempotent e belonging to H is calculated in the tensorproduct t.j",
"       of each irreducible with its contragredient representation.",
"    3) t.j(e) is decomposed in basismatrix <q>.j and projectionmatrix <p>.j.",
"",
NULL};

static proginfo_t pinfo =
   { "precond", "Precondensation", "$Revision: 1.16 $", helptext };




/* --------------------------------------------------------------------------
   Init() - Initialize everything

   Description:
     This function initializes global variables, processes command line
     options and arguments, and reads the .cfinfo and .tki files.

   Arguments:
     <argc>: Number of arguments
     <argv>: Command line arguments
   -------------------------------------------------------------------------- */

static void Init(int argc, char *argv[])

{
    int i;
    long fl1, nor1, noc1, fl2, nor2, noc2;
    char fn[LAT_MAXBASENAME+20];

    /* Initialize global data
       ---------------------- */
    memset(&TKInfo,0,sizeof(TKInfo));
    memset(TkiName,0,sizeof(TkiName));
    memset(CfIsLinked,0,sizeof(CfIsLinked));

    /* Initialize the MeatAxe library
       ------------------------------ */
    MtxInit();

    /* Process command line options
       ---------------------------- */
    MtxSetCommandLine(argc,argv,&pinfo);
    while ((i = MtxGetOpt("")) != OPT_END)
    {
	switch (i)
	{
	    case 's': opt_s = 1; break;
	    default:
		MtxErrorExit("Unknown option '%c'",i);
		break;
	}
    }

    /* Process remaining command line arguments
       ---------------------------------------- */
    if (opt_ind != argc - 3)
	MtxErrorExit("%s: %E\n",pinfo.name,ERR_BADUSAGE);
    strcpy(TkiName,argv[opt_ind]);

    /* Read info files
       --------------- */
    if (Lat_ReadInfo(&InfoM,argv[opt_ind + 1]) != 0)
	MtxErrorExit("Error reading %s.cfinfo",argv[opt_ind + 1]);
    if (Lat_ReadInfo(&InfoN,argv[opt_ind + 2]) != 0)
	MtxErrorExit("Error reading %s.cfinfo",argv[opt_ind + 2]);



    /* Initialize the TKInfo structure
       ------------------------------- */
    TKInfo.NGen = InfoM.NGen;
    TKInfo.NCf = 0;
    TKInfo.Order = 1;
    for (i = 0; i < LAT_MAXCF; ++i)
	TKInfo.CfIndex[0][i] = TKInfo.CfIndex[1][i] = -1;

    /* Some additional checks on input files
       ------------------------------------- */
    if (InfoM.Field != InfoN.Field)
    {
	MtxErrorExit("Incompatible representations: %s is over GF(%d), "
	   "%s is over GF(%d)",InfoM.BaseName,InfoM.Field,InfoN.BaseName,
	   InfoN.Field);
    }
    if (InfoM.NGen != InfoN.NGen)
    {
	MtxErrorExit("Incompatible representations: %s has %d generators, "
	   "%s has %d generators",InfoM.BaseName,InfoM.NGen,InfoN.BaseName,
	   InfoN.NGen);
    }

    /* Print start message
       ------------------- */
    sprintf(fn,"%s.1",InfoM.BaseName);
    fclose(zreadhdr(fn,&fl1,&nor1,&noc1));
    sprintf(fn,"%s.1",InfoN.BaseName);
    fclose(zreadhdr(fn,&fl2,&nor2,&noc2));
    MESSAGE(0,("Beginning pre-condensation of dimension %ld x %ld = %ld\n",
	nor1,nor2,nor1*nor2));
}



/* --------------------------------------------------------------------------
   IsDual() - Find matching constituent

   Description:
     This function decides is a given contituent of M is dual to a given 
     contituent of N.

   Arguments:
     <minfo>: Pointer to <CfInfo> structure for the constituent of M.
     <mgen>: Generators of the constituent of M.
     <nj>: Index of constituent of N.

   Return:
     1 = the constituents are dual, 0 = they are not dual.
   -------------------------------------------------------------------------- */

static int IsDual(CfInfo *minfo, matrix_t *mgen[], int nj)

{
    matrix_t *ngen[MAXGEN];	/* Generators of constituent */
    int j;
    int result;

    /* First check: Dimensions and splitting field must match
       ------------------------------------------------------ */
    if (InfoN.Cf[nj].dim != minfo->dim || InfoN.Cf[nj].spl != minfo->spl)
	return 0;

    /* Read the (contragrediate) generators and compare
       ------------------------------------------------ */
    MESSAGE(2,(" (%s%s)",InfoN.BaseName,Lat_CfName(&InfoN,nj)));
    Lat_ReadCfGens(&InfoN,nj,ngen,LAT_RG_INVERT|LAT_RG_TRANSPOSE|LAT_RG_STD);
    result = IsIsomorphic(InfoM.NGen,(const matrix_t **)mgen,minfo,
	(const matrix_t **)ngen,Trans + TKInfo.NCf,1);
    for (j = 0; j < InfoN.NGen; j++)
	MatFree(ngen[j]);

    return result;
}




/* --------------------------------------------------------------------------
   FindConstituentInN() - Find matching constituent

   Description:
     This function finds, for a given constituent of M, the corresponding
     constituent in N (if any exists). The generators in <mgen> are supposed
     to be in standard basis with respect to the constituent's idword.

   Arguments:
     <mj>: Index of the constituent of M.
     <mgen>: Generators of the constituent of M.

   Return:
     Index of corresponding constituent of N, or -1 if not found.
   -------------------------------------------------------------------------- */

static int FindConstituentInN(int mj, matrix_t *mgen[])

{
    int nj;

    /* Main loop: for each constituent of N
       ------------------------------------ */
    for (nj = 0; nj < InfoN.NCf; ++nj)
    {
	/* Discard constituents which are already linked
	   --------------------------------------------- */
	if (CfIsLinked[nj])
	   continue;

	/* Compare with <nj>-th constituent of N
	   ------------------------------------- */
	if (IsDual(InfoM.Cf + mj,mgen,nj))
	{
	    CfIsLinked[nj] = 1;	    /* Mark as linked */
	    return nj;
	}
    }

    /* Corresponding constituent was not found
       --------------------------------------- */
    return -1;
}



/* --------------------------------------------------------------------------
   MkEndo() - Calculate endomorphism ring

   Description:
     This function calculates, for a given irreducible constituent, a
     basis for the endomoprhism ring.

   Arguments:
     <ngen>: Number of generators
     <gen>: Generators.
     <m>: Constituent index
     <cf>: Pointer to constituent information.
     <endo>: Pointer to an array where the basis of End(V) is stored.
     <maxendo>: Maximum number of matrices that can be stored in <endo>.
     <basis>: The 'E-basis' is stored here
   -------------------------------------------------------------------------- */

static void MkEndo(int ngen, const matrix_t **gen, int m, const CfInfo *cf,
    matrix_t **endo, int maxendo)
{
    matrix_t *pw, *nsp;
    wgdata_t *wg;
    int i;
    MTXCONTEXT(MkEndo);

    MTX_VERIFY(maxendo >= cf->spl);

    /* Make the peak word kernel
       ------------------------- */
    wg = WgInit(ngen,gen);
    pw = WgMakeWord(wg,cf->peakword);
    WgFree(wg);
    nsp = MatNullSpace__(MatInsert(pw,cf->peakpol));
    MTX_ASSERT(nsp->nor == cf->spl);
    MatFree(pw);

    /* Calculate a basis of the the endomorphism ring
       ---------------------------------------------- */
    i = MakeEndomorphisms(ngen,gen,nsp,endo);
    MTX_ASSERT(i == 0);

    MatFree(nsp);
}




/* --------------------------------------------------------------------------
   MakeQ() - Calculate the Q matrix

   Description:
     This function calculates the embedding of the condensed tensor product
     (VxW)e into the tensor product VxW for one irreducible constituent.

   Arguments:
     <n>: Constituent index.
     <spl>: Splitting field degree.
     <endo>: Basis of endomorphism ring.
   -------------------------------------------------------------------------- */

static void MakeQ(int n, int spl, const matrix_t **endo)
{		  
    int i;
    int dim = endo[0]->nor;
    matrix_t *q = MatAlloc(endo[0]->fl,spl,dim*dim);
    char fn[200];
    for (i = 0; i < spl; ++i)
    {
	int j;
	matrix_t *y = MatInv(Trans[n]), *x;
	MatMul(y,endo[i]);
	x = MatTr(y);
	MatFree(y);
	for (j = 1; j <= dim; ++j)
	    MatCopyRegion(q,i+1,(j-1) * dim+1,x,j,1,1,-1);
	MatFree(x);
    }
    sprintf(fn,"%s.q.%d",TkiName,n+1);
    MESSAGE(2,("Writing %s\n",fn));
    MatSave(q,fn);
    MatFree(q);
}



/* --------------------------------------------------------------------------
   MakePQ() -

   Description:

   Arguments:
     <n>: Constituent index
     <mj>: Index of corresponding constituent of M.
     <nj>: Index of corresponding constituent of N.
   -------------------------------------------------------------------------- */

#define MAXENDO 20

static void MakePQ(int n, int mj, int nj)

{
    matrix_t *mgen[MAXGEN], *estar[MAXENDO], *endo[MAXENDO], *e, *ei;
    char fn[200];
    int dim = InfoM.Cf[mj].dim;
    int spl = InfoM.Cf[mj].spl;
    int i;
    matrix_t *p;
    PTR pptr;

    MESSAGE(1,("Condensing %s%s x ",InfoM.BaseName,Lat_CfName(&InfoM,mj)));
    MESSAGE(1,("%s%s, [E:k]=%d\n",InfoN.BaseName,Lat_CfName(&InfoN,nj),spl));

    /* Read the generators for the constituent of M
       -------------------------------------------- */
    Lat_ReadCfGens(&InfoM,mj,mgen,LAT_RG_STD);

    /* Make the endomorphism ring
       -------------------------- */
    MESSAGE(2,("Calculating endomorphism ring\n"));
    MkEndo(InfoM.NGen,(const matrix_t **)mgen,mj,InfoM.Cf + mj,endo,MAXENDO);

    /* Calculate the Q matrix
       ---------------------- */
    MESSAGE(2,("Calculating embedding of E\n"));
    MakeQ(n,spl,(const matrix_t **)endo);
    
    /* Calculate the E* matrices
       Note: We should use the symmetry under i<-->k here!
       --------------------------------------------------- */
    MESSAGE(2,("Calculating projection on E\n"));
    e = MatAlloc(zfl,spl,spl);
    pptr = e->d;
    for (i = 0; i < spl; ++i)
    {
	int k;
	for (k = 0; k < spl; ++k)
	{
	    FEL f;
	    matrix_t *x = MatDup(endo[i]);  
	    MatMul(x,endo[k]);
	    f = MatTrace(x);
	    zinsert(pptr,k+1,f);
	    MatFree(x);
	}
	zsetlen(spl);
	zadvance(&pptr,1);
    }
    ei = MatInv(e);
    MatFree(e);

    for (i = 0; i < spl; ++i)
    {
	int k;
	PTR p;
	estar[i] = MatAlloc(zfl,dim,dim);
	p = ei->d;
	zsetlen(spl);
	zadvance(&p,i);
	for (k = 0; k < spl; ++k)
	    MatAddMul(estar[i],endo[k],zextract(p,k+1));
    }
    MatFree(ei);

    /* Calculate the P matrix
       ---------------------- */
    p = MatAlloc(zfl,dim*dim,spl);
    pptr = p->d;
    for (i = 1; i <= dim; ++i)
    {
	int j;
	for (j = 1; j <= dim; ++j)
	{
	    int r;
	    matrix_t *x = MatAlloc(zfl,dim,dim);
	    MatCopyRegion(x,1,i,Trans[n],1,j,dim,1);
	    for (r = 0; r < spl; ++r)
	    {
		FEL f;
		matrix_t *z = MatDup(x);
		MatMul(z,estar[r]);
		f = MatTrace(z);
		zinsert(pptr,r+1,f);
		MatFree(z);
	    }
	    MatFree(x);
	    zsetlen(spl);
	    zadvance(&pptr,1);
	}
    }

    sprintf(fn,"%s.p.%d",TkiName,n+1);
    MESSAGE(2,("Writing %s\n",fn));
    MatSave(p,fn);

    /* Clean up
       -------- */
    MatFree(p);
    for (i = 0; i < spl; ++i)
        MatFree(endo[i]);
}




/* --------------------------------------------------------------------------
   CalcDim() - Calculate dimension of condensed module

   Description:
     This function calculates the dimension of the condensed tensor product.

   Remarks:
     The dimension is stored into <TKInfo.Dim>.
   -------------------------------------------------------------------------- */

static void CalcDim()

{
    int i;

    TKInfo.Dim = 0;
    for (i = 0; i < TKInfo.NCf; ++i)
    {
	int m = TKInfo.CfIndex[0][i];
	int n = TKInfo.CfIndex[1][i];
	TKInfo.Dim += InfoM.Cf[m].mult * InfoN.Cf[n].mult * InfoM.Cf[m].spl;
    }
    MESSAGE(0,("Dimension of condensed module = %d\n",TKInfo.Dim));
}


/* --------------------------------------------------------------------------
   main() - Program entry point
   -------------------------------------------------------------------------- */

int main(int argc, char *argv[])

{
    int mj;

    Init(argc,argv);			/* Initialize everything */

    /* Main loop: for all constituents of M
       ------------------------------------ */
    for (mj = 0; mj < InfoM.NCf; mj++)
    {
	matrix_t *mgen[MAXGEN];	    /* Generators for const. of M */
	int k;
	int nj;

	MESSAGE(0,("Examining %s%s ...",InfoM.BaseName,Lat_CfName(&InfoM,mj)));

	/* Read the generators for the <mj>-th contituent of M, and find
	   the corresponting (=contragredient) constituent in N
	   ------------------------------------------------------------- */
	Lat_ReadCfGens(&InfoM,mj,mgen,LAT_RG_STD);
	nj = FindConstituentInN(mj,mgen);

	/* Calculate the P and Q matrix for this constituent
	   ------------------------------------------------- */
	if (nj >= 0)
	{
	    MESSAGE(0,(" ... corresponds to %s%s\n",InfoN.BaseName,
		Lat_CfName(&InfoN,nj)));
	    TKInfo.CfIndex[0][TKInfo.NCf] = mj;
	    TKInfo.CfIndex[1][TKInfo.NCf] = nj;
	    MakePQ(TKInfo.NCf,mj,nj);
	    TKInfo.NCf++;
	}
	else
	    MESSAGE(0,(" not found in %s\n",InfoN.BaseName));

	/* Clean up
	   -------- */
	for (k = 0; k < InfoM.NGen; ++k)
	    MatFree(mgen[k]);
    }

    CalcDim();				/* Calculate dimension */
    TK_WriteInfo(&TKInfo,TkiName);	/* Write .tki file */
    MtxCleanup();
    return 0;
}
